from pydantic import ConfigDict, Field
from typing import Union, Annotated, Literal, List
import matplotlib.pyplot as plt
import numpy as np
from .. import base_model as bm
from ..plot_utils import convert_fig_to_html_img

class SubassemblyAlignmentV0(bm.ConstructionBase):
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                "component": "PS0001",
                "name": "subassembly alignment",
                "measurement_date": "2023-01-01T12:00:00+01:00",
                "location": "BU",
                "user_created": "hayden",
                "version": "v0",
                "target": [639.141118, 287.244992, 64.009534,-0.048954],
                "actual": [639.141118, 287.244992, 64.009534,-0.048954],
                "delta":  [639.141118, 287.244992, 64.009534,-0.048954]
            },
        ],
        'table': 'assembly',
        'component_types': ['Production Subassembly', 'PreProduction Subassembly'],
        'module_types': [],
        'description': 'An assembly measurement for the relative alignment between the centers of the LGAD and ETROC'
    })
    name: Literal['subassembly alignment']
    version: Literal["v0"] = "v0"

    target: List[float]
    actual: List[float]
    delta: List[float]

    def plot(self):
        """
        A vector like [dX (um), dY (um), dZ (um), dRot (degrees)]
        """
        assy_data = np.array(self.delta)
        fig, ax = plt.subplots(figsize=(4, 4))
        # add your plot command here
        ax.plot(assy_data[0]*1000, assy_data[1]*1000, 'o') # 'o' specifies points
        ax.set_xlabel('dx')
        ax.set_ylabel('dy')
        ax.grid()
        #have these be the default values and if the point goes outside it just zooms out with some wiggle room
        ax.set_xlim([-200, 200])
        ax.set_ylim([-200, 200])
        ax.set_title(f'ETROC+LGAD Alignment (um)') 
        
        ax.axhline(0, color='black', linewidth=1)
        ax.axvline(0, color='black', linewidth=1)
        return fig
    
    def html_display(self):
        fig = self.plot()
        return convert_fig_to_html_img(fig)
    
SubassemblyAlignmentType = Annotated[Union[SubassemblyAlignmentV0], Field(discriminator="version")]


from typing import Literal, List, Annotated, Union
from pydantic import ConfigDict, field_validator, Field
import numpy as np
import matplotlib.pyplot as plt
from ..plot_utils import convert_fig_to_html_img
from .. import base_model

class PickAndPlaceV0(base_model.ConstructionBase):
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                "module": "PBU0001",
                "component": "PE0001",
                "component_pos": 1, 
                "version": "v0",
                "name": "pick and place survey",
                "measurement_date": "2023-01-01T12:00:00+01:00",
                "location": "BU",
                "user_created": "hayden",
                "stage": "precure",
                "target": [639.141118, 287.244992, 64.009534,-0.048954],
                "actual": [639.141118, 287.244992, 64.009534,-0.048954],
                "delta":  [639.141118, 287.244992, 64.009534,-0.048954]
            },
            {
                "module": "PBU0001",
                "component": "PE0001",
                "component_pos": 1, 
                "version": "v0",
                "name": "pick and place survey",
                "measurement_date": "2023-01-01T12:00:00+01:00",
                "stage": "precure",
                "location": "BU",
                "user_created": "hayden",
                "stage": "postcure",
                "target": [639.141118, 287.244992, 64.009534,-0.048954],
                "actual": [639.141118, 287.244992, 64.009534,-0.048954],
                "delta":  [639.141118, 287.244992, 64.009534,-0.048954]
            }
        ],
        'table': 'assembly',
        'component_types': ['Production Subassembly', 'PreProduction Subassembly', 'Ghost ETROC'],
        'module_types': ['PreProduction', 'Production', 'Digital'],
        'description': 'Gantry assembly step to measure the position of the subassemblies on the module (either before curing or after curing)'
    })
    name: Literal['pick and place survey']
    version: Literal["v0"] = "v0"

    stage: Literal["precure", "postcure"]
    target: List[float]
    actual: List[float]
    delta: List[float]

    @field_validator('target', 'actual', 'delta', mode="after")
    @classmethod
    def length_check(cls, v):
        if len(v) != 4:
            raise ValueError("The required length is 4 for target, actual and delta. It is [x, y, z, rot]")
        return v

    def plot(self):
        """
        A vector like [dX (um), dY (um), dZ (um), dRot (degrees)]
        """
        assy_data = np.array(self.delta)
        fig, ax = plt.subplots(figsize=(4, 4))
        # add your plot command here
        ax.plot(assy_data[0]*1000, assy_data[1]*1000, 'o') # 'o' specifies points
        ax.set_xlabel('dx')
        ax.set_ylabel('dy')
        ax.grid()

        ax.set_title(f'Pick and Place Alignment (um)') 
        ax.set_xlim([-20, 20])
        ax.set_ylim([-20, 20])                

        ax.axhline(0, color='black', linewidth=1)
        ax.axvline(0, color='black', linewidth=1)
        return fig
    
    def html_display(self):
        fig = self.plot()
        return convert_fig_to_html_img(fig)
    
PickAndPlaceType = Annotated[Union[PickAndPlaceV0], Field(discriminator="version")]



from pydantic import AfterValidator
from typing import List, Annotated
import numpy as np

def _pix_arr_shape(value: List[List[int]]) -> List[List[int]]:
    v_arr = np.array(value)
    if v_arr.shape != (16,16):
        raise ValueError(f"Your array is not the correct shape, it should be 16x16, you gave: {v_arr.shape}")
    return value

PixArr = Annotated[
    List[List[int]],
    AfterValidator(_pix_arr_shape),
]


from pydantic import ConfigDict, field_validator, Field
from typing import Literal, Optional
from typing import Literal, Optional, Annotated, Union
from .. import jinja_env
from pathlib import Path
from io import BytesIO
import base64
from .. import base_model as bm

try:
    from application import the_config
    from webdav3.client import Client #for cloud talking webdav is old but it works for CERNbox
    CERNBOX_CLIENT_DETAILS = {
        'webdav_hostname': the_config.WEBDAV_HOSTNAME,
        'webdav_login': the_config.WEBDAV_LOGIN,
        'webdav_password': the_config.WEBDAV_PASSWORD
    }
except ModuleNotFoundError as e:
    CERNBOX_CLIENT_DETAILS = None

class ImageTestV0(bm.ConstructionBase):
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                "component": "ET2.01-PT-NH21",
                "name": "IV Scan Image",
                "measurement_date": "2023-01-01T12:00:00+01:00",
                "location": "FL",
                "user_created": "hswanson",
                "version": "v0",
                "cernbox_img_path": "/eos/project/e/etl-construction-app-storage/TedTestResults/IV_Curve/IV_second_PT_NH_19_20_21_22.png",
                "comment": "this result was goods",
                "bias": 123
            }
        ],
        'table': 'test',
        'component_types': ['Prototype Hybrid'],
        'module_types': [],
        'description': 'IV curves from teds presentation: https://indico.cern.ch/event/1566730/contributions/6599657/attachments/3099619/5491873/Summary-of-bump-bonding-study-July2025.pdf'
    })
    name: Literal['IV Scan Image', "Baseline and Noisewidth Histogram Image", "Baseline and Noisewidth Image", "Time Resolution - Bias Voltage Scan Image"]
    version: Literal["v0"] = "v0"
    component: str

    cernbox_img_path: str
    comment: str
    bias: Optional[int] = None

    @field_validator("cernbox_img_path")
    @classmethod
    def check_img_exists(cls, v):
        if CERNBOX_CLIENT_DETAILS is None:
            return v
        v_path = Path(v)
        cernbox_relative_path = v_path.relative_to("/eos/project/e/etl-construction-app-storage/")
        client = Client(CERNBOX_CLIENT_DETAILS)
        if not client.check(str(cernbox_relative_path)):
            raise ValueError(f"This file ({cernbox_relative_path}) does not exist in the cernbox. The cernbox is the one for the web app, contact an admin to upload.")
        return v
    
    def plot(self):
        return None

    def html_display(self) -> str:
        if CERNBOX_CLIENT_DETAILS is None:
            print("This feature only works on the web application, returning nothing.")
            return None
        
        if self.cernbox_img_path is not None:
            # the path should be obtained by copying the "fuse path" on cernbox
            eos_fuse_path = Path(self.cernbox_img_path)
            cernbox_relative_path = str(
                eos_fuse_path.relative_to("/eos/project/e/etl-construction-app-storage/"))
            
            client = Client(CERNBOX_CLIENT_DETAILS)
            buff = BytesIO()
            client.download_from(
                remote_path = cernbox_relative_path, 
                buff = buff)
            buff.seek(0)
            encoded_img = base64.b64encode(buff.getvalue()).decode('utf-8')
            img_html = f'<img src="data:image/png;base64,{encoded_img}" alt="IV Scan Image">'
        else:
            img_html = "Resource image path not provided in test data"

        template = jinja_env.get_template('test_by_image.html')
        display_data = {
            "cernbox_img_path": self.cernbox_img_path,
            "comment": self.comment,
            "bias": self.bias
        }
        return template.render(
            display_data = display_data,
            img_html = img_html
        )
    
ImageTestType = Annotated[Union[ImageTestV0], Field(discriminator="version")]


from typing import Union, Optional
from pydantic import BaseModel, field_validator, model_validator, AwareDatetime, ConfigDict, PastDate
from datetime import datetime, date
import pytz
from .example_formatter import ExampleFormatterMixin
    
class ConstructionBase(BaseModel, ExampleFormatterMixin):
    model_config = ConfigDict(extra="forbid") #forbid any extra fields

    measurement_date: Union[PastDate, AwareDatetime] #force times to have timezone
    location: str
    user_created: str

    # make all 3 optional but enforce atleast module or component
    # in db, you can just enforce by the expected component type and module type
    module: Optional[str] = None
    component: Optional[str] = None
    component_pos: Optional[int] = None
    
    @field_validator('*')
    @classmethod
    def empty_str_to_none(cls, v):
        if isinstance(v, str) and v.strip() == '':
            return None
        return v
    
    @model_validator(mode='after')
    def enforce_part(self):
        """Makes sure a part is given, either module or component"""
        if not (self.module or self.component or self.component_pos):
            raise ValueError("Please specify a part")
        elif self.component_pos and not self.module:
            raise ValueError("Component position alone is not enough to identify a part for a test.")
        return self
    
    @classmethod
    def get_subclasses(cls):
        return tuple(cls.__subclasses__())

    @field_validator("measurement_date")
    @classmethod
    def validate_measurement_date(cls, v):
        if type(v) is date: #cannot do isinstance to differentiate between datetime and date objects!
            return v
        if not type(v) is datetime:
            raise ValueError(f"Inputted datetime is not a datetime object it is {type(v)}")
        if v.tzinfo is None:
            #need to do this otherwise it has a default for no timezone given!
            raise ValueError("Measurement data has no time zone information, see https://en.wikipedia.org/wiki/ISO_8601")
        
        #if measurement date has tz information, convert from that TZ to UTC time!
        return v.astimezone(pytz.utc) 
    
    def get_schema_val(self, key):
        """
        Gets you anything in the json schema, could be description, module_types, etc...
        """
        json_schema = self.model_config.get("json_schema_extra")
        return json_schema.get(key)
    
    


from pydantic import ConfigDict, Field, AliasChoices, model_validator
from typing import Literal, List, Union, Annotated
import matplotlib.pyplot as plt

from ..plot_utils import convert_fig_to_html_img
from .. import base_model as bm
from .. import jinja_env

class CurrentStabilityV0(bm.ConstructionBase):
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                "component": "FBK_LF1_ROL_054",
                "name": "Sensor Current Stability",
                "measurement_date": "2023-01-01T12:00:00+01:00",
                "location": "Universita e INFN Torino",
                "user_created": "fsiviero",
                "version": "v0",
                "current_stability": "A",
                "current": [1,2,3,4,5],
                "time": [1,2,3,4,5]
            }
        ],
        'table': 'test',
        'component_types': ['Prototype LGAD'],
        'module_types': [],
        'description': 'LGAD current stability test'
    })

    name: Literal['Sensor Current Stability'] = 'Sensor Current Stability'
    version: Literal["v0"] = "v0"
    
    # Inline the data fields from CurrentStabilityData
    current_stability: Union[None, Literal["A", "B", "C"]]
    current: Union[None, List[float]] = Field(validation_alias=AliasChoices('current','Current'))
    time: Union[None, List[float]] = Field(validation_alias=AliasChoices('time','Time'))

    @model_validator(mode='after')
    def same_lengths(self):
        if self.current is not None and self.time is not None:
            if len(self.current) != len(self.time):
                raise ValueError(f'Current and Time arrays should have the same lengths. Length of Current, Length of Time = ({len(self.current)}, {len(self.time)})')
        return self

    def plot(self):
        fig, ax = plt.subplots(figsize=(10, 6))
        if self.current is not None and self.time is not None:
            ax.plot(self.time, self.current, marker='o', color='deeppink')
            ax.set_xlabel('Time')
            ax.set_ylabel('Current')
            ax.set_title(f'Time vs Current')
            ax.grid(True)
        return fig

    def html_display(self):
        display_data = {
            "current_stability": self.current_stability
        }
        fig = self.plot()

        template = jinja_env.get_template('sensors_plot.html')
        return template.render(
            plot = convert_fig_to_html_img(fig),
            display_data = display_data
        )

CurrentStabilityType = Annotated[Union[CurrentStabilityV0], Field(discriminator="version")]


from pydantic import ConfigDict, Field, AliasChoices, model_validator
from typing import Literal, List, Union, Annotated
import matplotlib.pyplot as plt

from ..plot_utils import convert_fig_to_html_img
from .. import base_model
from .. import jinja_env  # relative to the top-level package "etlup"

class ChargeCollectionV0(base_model.ConstructionBase):
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                "component": "FBK_LF1_ROL_054",
                "name": "Sensor Charge Collection",
                "measurement_date": "2023-01-01T12:00:00+01:00",
                "location": "UT",
                "user_created": "fsiviero",
                "version": "v0",
                "irradiation_level": 1E15,
                "side": "B", 
                "geometry": "1x2 PIN",
                "measuring_temperature": -20,
                "charge": [1,2,3,4,5],
                "voltage": [1,2,3,4,5],
            }
        ],
        'table': 'test',
        'component_types': ['Prototype LGAD'],
        'module_types': [],
        'description': 'Tests for QA-QC test structures'
    })

    name: Literal['Sensor Charge Collection'] = 'Sensor Charge Collection'
    version: Literal["v0"] = "v0"
    
    # Inline the data fields from ChargeCollectionData
    side: Union[None, Literal["A", "B"]]      = Field(validation_alias=AliasChoices('side','Side'))
    geometry: Union[None, Literal["1x1 LGAD", 
                                  "1x1 PIN", 
                                  "1x2 PIN"]] = Field(validation_alias=AliasChoices('geometry','Geometry'))
    irradiation_level: Union[None, float]     = Field(validation_alias=AliasChoices('irradiation_level','Irradiation Level'))
    measuring_temperature: Union[None, float] = Field(validation_alias=AliasChoices('measuring_temperature','Measuring Temperature'))
    voltage: Union[None, List[float]]         = Field(validation_alias=AliasChoices('voltage','Voltage'))
    charge: Union[None, List[float]]          = Field(validation_alias=AliasChoices('charge','Charge'))

    @model_validator(mode='after')
    def same_lengths(self):
        if self.charge is not None and self.voltage is not None:
            if len(self.charge) != len(self.voltage):
                raise ValueError(f'Charge and Voltage arrays should have the same lengths. Length of Charge, Length of Voltage = ({len(self.charge)}, {len(self.voltage)})')
        return self

    def plot(self):
        fig, ax = plt.subplots(figsize=(10, 6))
        if self.voltage is not None and self.charge is not None:
            ax.plot(self.voltage, self.charge, marker='o', color='green')
            ax.set_xlabel('Voltage')
            ax.set_ylabel('Charge')
            ax.set_title(f'Charge vs Voltage')
            ax.grid(True)
        return fig

    def html_display(self):
        display_data = {
            "side": self.side,
            "geometry": self.geometry,
            "irradiation_level": self.irradiation_level,
            "measuring_temperature": self.measuring_temperature
        }
        fig = self.plot()

        template = jinja_env.get_template('sensors_plot.html')
        return template.render(
            plot = convert_fig_to_html_img(fig),
            display_data = display_data
        )

# class ChargeCollectionV1(ChargeCollectionV0):
#     version: Literal['v1'] = "v1"
#     boop: str

ChargeCollectionType = Annotated[Union[ChargeCollectionV0], Field(discriminator="version")]



from pydantic import ConfigDict, Field, AliasChoices, model_validator
from typing import Literal, List, Union, Annotated
import matplotlib.pyplot as plt

from ..plot_utils import convert_fig_to_html_img
from .. import base_model, jinja_env

class MPVStabilityV0(base_model.ConstructionBase):
    model_config = ConfigDict(json_schema_extra={
        'examples': [
            {
                "component": "FBK_LF1_ROL_054",
                "name": "Sensor MPV Stability",
                "measurement_date": "2023-01-01T12:00:00+01:00",
                "location": "UT",
                "user_created": "fsiviero",
                "version": "v0",
                "side": "B", 
                "geometry": "1x1 LGAD",
                "mpv": [1,2,3,4,5],
                "time": [1,2,3,4,5],
                "mpv_stability": 0.2
            }
        ],
        'table': 'test',
        'component_types': ['Prototype LGAD'],
        'module_types': [],
        'description': 'Tests for QA-QC test structures'
    })

    name: Literal['Sensor MPV Stability'] = 'Sensor MPV Stability'
    version: Literal["v0"] = "v0"
    
    # Inline the data fields from MPVStabilityData
    side: Union[None, Literal["A", "B"]] = Field(validation_alias=AliasChoices('side','Side'))
    geometry: Union[None, Literal["1x1 LGAD", "1x1 PIN", "1x2 PIN"]] = Field(validation_alias=AliasChoices('geometry','Geometry'))
    mpv: Union[None, List[float]] = Field(validation_alias=AliasChoices('mpv','MPV'))
    time: Union[None, List[float]] = Field(validation_alias=AliasChoices('time','Time'))
    mpv_stability: Union[None, float] = Field(validation_alias=AliasChoices('mpv_stability','MPV Stability'))
    
    @model_validator(mode='after')
    def same_lengths(self):
        if self.mpv is not None and self.time is not None:
            if len(self.mpv) != len(self.time):
                raise ValueError(f'MPV and Time arrays should have the same lengths. Length of MPV, Length of Time = ({len(self.mpv)}, {len(self.time)})')
        return self

    def plot(self):
        fig, ax = plt.subplots(figsize=(10, 6))
        if self.time is not None and self.mpv is not None:
            ax.plot(self.time, self.mpv, marker='o', color='black')
            ax.set_xlabel('Time')
            ax.set_ylabel('MPV')
            ax.set_title(f'MPV Stability')
            ax.grid(True)
        return fig

    def html_display(self):
        display_data = {
            "side": self.side,
            "geometry": self.geometry,
            "mpv_stability": self.mpv_stability
        }
        fig = self.plot()

        template = jinja_env.get_template('sensors_plot.html')
        return template.render(
            plot = convert_fig_to_html_img(fig),
            display_data = display_data
        )

MPVStabilityType = Annotated[Union[MPVStabilityV0], Field(discriminator="version")]